#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyGoo - An Object-Graph mapper
# Copyright (c) 2010 Nicolas Wack <wackou@gmail.com>
#
# PyGoo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyGoo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from pygoo.abstractdirectedgraph import AbstractDirectedGraph, Equal
from pygoo.baseobject import BaseObject, get_node
from pygoo.utils import reverse_lookup
from pygoo import ontology
import types
import logging

log = logging.getLogger('pygoo.ObjectGraph')



def wrap_node(node, node_class = None):
    if node_class is None:
        node_class = BaseObject
    return node_class(basenode = node)

def unwrap_node(node):
    node_class = node.__class__ if isinstance(node, BaseObject) else None
    node = get_node(node)

    return node, node_class



class ObjectGraph(AbstractDirectedGraph):
    """An ObjectGraph is an undirected graph of ObjectNodes.
    An ObjectNode "looks like" an object, with a class type and any number of properties/attributes,
    which can be either literal values or other objects in the graph. These properties are accessed
    using dotted attribute notation.

    The ObjectGraph edges (links) are a bit special in that they are undirected but have a different
    name depending on the direction in which they are followed, e.g.: if you have two nodes, one
    which represent a series, and another one which represent an episode, the link will be named
    "series" when going from the episode to the series, but will be named "episodes" when going from
    the series to the episode.

    When setting a link in the graph, you should define a name for the reverse name of the link, but if
    you don't a default name of "is(%Property)Of" will be given to it.
    For instance, if we have movie.director == personX, we should also have
    personX.isDirectorOf == movie (actually "movie in personX.isDirectorOf").

    The ObjectGraph class provides ways of querying objects in it using type information,
    properties matching filters, or just plain lambda functions that returns whether a node
    is acceptable or not.

    An ObjectGraph can be thought of as a context in which objects live.

    Even though the dotted attribute access makes this less visible, the links between ObjectNodes
    are first-class citizens also, and can themselves have attributes, such as confidence, etc...
    """

    # this should be set to the used ObjectNode class (ie: MemoryObjectNode or Neo4jObjectNode)
    # in the corresponding derived ObjectGraph class
    _object_node_class = type(None)

    def __init__(self, dynamic = False):
        """Creates an ObjectGraph.

        - if dynamic = True, we have static inheritance (valid classes need to be set explicitly)
        - if dynamic = False, we have dynamic type classes (valid classes are automatically updated if the object has the correct properties)
        """
        ontology.register_graph(self)
        self._dynamic = dynamic

    def revalidate_objects(self):
        if not self._dynamic:
            return

        log.info('revalidating objects in graph %s' % self)
        for node in self.nodes():
            node.update_valid_classes()

    def __getattr__(self, name):
        # if attr is not found and starts with an upper case letter, it might be the name
        # of one of the registered classes. In that case, return a function that would instantiate
        # such an object in this graph
        if name[0].isupper() and name in ontology.class_names():
            def inst(basenode = None, **kwargs):
                # if we're copying a node from a different graph, we need to intercept it here to
                # add it correctly with its dependencies instead of creating from scratch
                if basenode is not None and basenode.node.graph() != self:
                    return self.add_object(basenode)

                return ontology.get_class(name)(basenode = basenode, graph = self, **kwargs)

            return inst

        raise AttributeError, name

    def __contains__(self, node):
        """Return whether this graph contains the given node  or object (identity)."""
        return self.contains(get_node(node))


    def add_link(self, node, name, other_node, reverse_name):
        # otherNode should always be a valid node
        self.add_directed_edge(node, name, other_node)
        self.add_directed_edge(other_node, reverse_name, node)

    def remove_link(self, node, name, other_node, reverse_name):
        # other_node should always be a valid node
        self.remove_directed_edge(node, name, other_node)
        self.remove_directed_edge(other_node, reverse_name, node)


    def add_node(self, node, recurse = Equal.OnIdentity, excluded_deps = list()):
        return self.add_object(BaseObject(node), recurse, excluded_deps)


    def add_object(self, node, recurse = Equal.OnIdentity, excluded_deps = list()):
        """Add an object and its underlying node and its links recursively into the graph.

        If some dependencies of the node are already in the graph, we should not add
        new instances of them but use the ones already there (ie: merge links).

        This strategy should be configurable, and offer at least the following choices:
          - recurse = OnIdentity   : do not add the dependency only if the exact same node is already there
          - recurse = OnValue      : do not add the dependency only if there is already a node with the exact same properties
          - recurse = OnValidValue : do not add the dependency only if there is already a node with the same valid properties
          - recurse = OnUnique     : do not add the dependency only if there is already a node with the same unique properties
        """
        # FIXME: not necessarily correct, but safer for now to avoid infinite recursion
        #        ie, if we add a node without a class, we won't know its implicit dependencies
        node = node.node.virtual()
        log.debug('Adding to graph: %s - node: %s' % (self, node))
        node, node_class = unwrap_node(node)

        if node_class is None:
            raise TypeError("Can only add BaseObjects to a graph at the moment...")

        # first make sure the node's not already in the graph, using the requested equality comparison
        # TODO: if node is already there, we need to decide what to do with the additional information we have
        #       in the added node dependencies: update missing properties, update all properties (even if already present),
        #       update non-valid properties, ignore new data, etc...
        excluded_properties = node_class.schema._implicit if node_class is not None else []
        log.debug('exclude properties: %s' % excluded_properties)

        gnode = self.find_node(node, recurse, excluded_properties)
        if gnode is not None:
            return wrap_node(gnode, node_class)

        # if node isn't already in graph, we need to make a copy of it that lives in this graph

        # first import any other node this node might depend on
        newprops = []
        for prop, value, reverse_name in reverse_lookup(node, node_class):
            #if (isinstance(value, AbstractNode) or
            #    (isinstance(value, list) and isinstance(value[0], AbstractNode))):
            if isinstance(value, types.GeneratorType):
                # use only the explicit properties here
                if prop not in excluded_properties and value not in excluded_deps:
                    imported_nodes = []
                    for v in value:
                        log.debug('Importing dependency %s: %s' % (prop, v))
                        imported_nodes.append(self.add_object(wrap_node(v, node_class.schema.get(prop)),
                                                              recurse,
                                                              excluded_deps = excluded_deps + [node]).node)
                    newprops.append((prop, imported_nodes, reverse_name))
            else:
                newprops.append((prop, value, reverse_name))

        # actually create the node
        result = self.create_node(newprops, _classes = node._classes)

        return wrap_node(result, node_class)


    def __iadd__(self, node):
        """Should allow node, but also list of nodes, graph, ..."""
        if isinstance(node, list):
            for n in node:
                self.add_object(n)
        else:
            self.add_object(node)

        return self


    ### Search methods

    def find_node(self, node, cmp = Equal.OnIdentity, exclude_properties = list()):
        """Return a node in the graph that is equal to the given one using the
        specified comparison type, or None if not found."""

        # TODO: could prefilter all the nodes with their classes, instead of comparing
        #       properties for nodes which are not even the same classes, that would surely
        #       make things much much faster, especially if classes are stored in a tuple

        if cmp == Equal.OnIdentity:
            if self.contains(node):
                log.debug('%s already in graph %s (id)...' % (node, self))
                return node

        elif cmp == Equal.OnValue:
            for n in self.nodes():
                if node.same_properties(n, exclude = exclude_properties):
                    log.debug('%s already in graph %s (value)...' % (node, self))
                    return n

        elif cmp == Equal.OnLiterals:
            for n in self.nodes():
                if node.same_properties(n, n.literal_keys(), exclude = exclude_properties):
                    log.debug('%s already in graph %s (literals)...' % (node, self))
                    return n

        elif cmp == Equal.OnUnique:
            obj = node.virtual()
            props = list(set(obj.unique_properties()) - set(exclude_properties))
            #print 'find_node OnUnique:', node
            #print 'Comparing on props:', props
            for n in self.nodes():
                #print '  ', n
                if node.same_properties(n, props, cmp = Equal.OnUnique):
                    log.debug('%s already in graph %s (unique)...' % (node, self))
                    return n

        else:
            raise NotImplementedError

        return None

    def find_all(self, node_type = None, valid_node = lambda x: True, **kwargs):
        """This method returns a list of the objects of the given type in this graph for which
        the cond function returns True (or sth that evaluates to True).
        It will also only keep those objects that have properties which match the given keyword
        args dictionary.

        When using both the cond function and the type argument, it is useful to know that the
        type is checked first, so that the cond function can safely assume that only objects of
        the correct type are given to it.

        When using keyword args for filtering, you can chain properties using '_' between them.
        it should be configurable whether property value matching should be case-insensitive or
        use regexps in the case of strings.

        If no match is found, it returns an empty list.

        examples:
          g.find_all(node_type = Movie)
          g.find_all(Episode, lambda x: x.season = 2)
          g.find_all(Movie, lambda m: m.releaseYear > 2000)
          g.find_all(Person, role_movie_title = 'The Dark Knight')
          g.find_all(Character, isCharacterOf_movie_title = 'Fear and loathing.*', regexp = True)
        """
        return list(self._find_all(node_type, valid_node, **kwargs))


    def _find_all(self, node_type = None, valid_node = lambda x: True, **kwargs):
        """Implementation of findAll that returns a generator."""
        if isinstance(node_type, basestring):
            node_type = ontology.get_class(node_type)

        for node in self.nodes_from_class(node_type) if node_type else self.nodes():
            # TODO: should this go before or after the properties checking? Which is faster in general?
            try:
                if not valid_node(node):
                    continue
            except:
                continue

            valid = True
            for prop, value in kwargs.items():
                try:
                    # FIXME: this doesn't work with lists of objects
                    if isinstance(value, BaseObject):
                        value = value.node

                    if node.get_chained_properties(prop.split('_')) != value:
                        valid = False
                        break
                except AttributeError:
                    valid = False
                    break

            if not valid:
                continue

            if node_type:
                yield node_type(node)
            else:
                yield node


    def find_one(self, node_type = None, valid_node = lambda x: True, **kwargs):
        """Returns a single result. see findAll for description.
        Raises an exception if no result was found."""
        # NB: as _findAll is a generator, this should be fairly optimized
        result = self._find_all(node_type, valid_node, **kwargs)
        try:
            return result.next()
        except StopIteration:
            raise ValueError('Could not find given %s with props %s' % (node_type.__name__, kwargs))

    def find_or_create(self, node_type, **kwargs):
        '''This method returns the first object in this graph which has the specified type and
        properties which match the given keyword args dictionary.
        If no match is found, it creates a new object with the keyword args, inserts it in the
        graph, and returns it.

        example: g.findOrCreate(Series, title = 'Arrested Development')'''
        try:
            return self.find_one(node_type, **kwargs)
        except ValueError:
            return node_type(graph = self, **kwargs)

    ## Let's make PyLint happy and let it know that the following methods are still not implemented

    def clear(self):
        """Delete all nodes and links in this graph."""
        raise NotImplementedError

    def create_node(self, props = list(), _classes = set()):
        """Create a node with the given properties and classes."""
        raise NotImplementedError

    def delete_node(self, node):
        """Remove a given node from this graph.

        Strategies for what to do with linked nodes should be configurable, ie:
        remove incoming/outgoing linked nodes as well, only remove link but do
        not touch linked nodes, etc..."""
        raise NotImplementedError

    def nodes(self):
        """Return an iterable on all the nodes in the graph."""
        raise NotImplementedError

    def nodes_from_class(self, cls):
        """Return an iterable on the nodes of a given class."""
        raise NotImplementedError

    def contains(self, node):
        """Return whether this graph contains the given node.

        multiple strategies can be used here for determing object equality, such
        as all properties equal, the primary subset of properties equal, etc...
        (those are defined by the ObjectNode)"""
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyGoo - An Object-Graph mapper
# Copyright (c) 2010 Nicolas Wack <wackou@gmail.com>
#
# PyGoo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyGoo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from pygootest import *

class TestAdvancedGraph(TestCase):

    def setUp(self):
        ontology.reload_saved_ontology('media')

    def testOnUniqueComparison(self):
        # mimics a typical import scenario for smewt:
        #  - we first import a few files
        #  - then we reimport them, with the same values.
        #  - graph should be the same, no duplicates
        ontology.import_all_classes()

        # initial import
        tmp1 = MemoryObjectGraph()
        s = tmp1.Series(title = 'Monk')
        ep = tmp1.Episode(series = s, season = 1, episodeNumber = 1, title = 'ep 1x01')
        epfile = tmp1.Media(metadata = ep, filename = 'ep_1x01.avi')

        sub = tmp1.Subtitle(metadata = ep, language = 'en')
        subfile = tmp1.Media(metadata = sub, filename = 'ep_1x01.en.srt')

        #tmp1.display_graph()

        collection = MemoryObjectGraph()
        epfilec = collection.add_object(epfile, recurse = Equal.OnUnique)

        print epfile.node.graph, epfile
        print epfilec.node.graph, epfilec

        print epfilec.metadata

        print 'from collec', collection.find_node(epfilec.node, cmp = Equal.OnUnique)
        #epfilec2 = collection.Media(metadata = epfilec.metadata, filename = 'ep_1x01.avi', other = 'different, but same unique values')
        #collection.display_graph()
        #n1 = collection.find_node(epfilec.node, cmp = Equal.OnUnique)
        #n2 = collection.find_node(epfilec2.node, cmp = Equal.OnUnique)
        #n3 = collection.find_node(epfile.node, cmp = Equal.OnUnique)

        #print 'from collec copy', n1
        #print 'from collec copy2', n2
        #print 'from outside', n3

        #self.assertEqual(n1, n2)
        #self.assert_(n3 is not None)

        #collection.display_graph()

        collection.add_object(subfile, recurse = Equal.OnUnique)

        print 'collec 1'
        #collection.display_graph()

        self.assertEqual(len(collection.find_all(Episode)), 1)
        self.assertEqual(len(collection.find_all(Subtitle)), 1)
        self.assertEqual(len(collection.find_all(Media)), 2)

        # second import, same objects
        tmp2 = MemoryObjectGraph()
        s = tmp2.Series(title = 'Monk')
        ep = tmp2.Episode(series = s, season = 1, episodeNumber = 1, title = 'ep 1x01')
        epfile = tmp2.Media(metadata = ep, filename = 'ep_1x01.avi')

        sub = tmp2.Subtitle(metadata = ep, language = 'en')
        subfile = tmp2.Media(metadata = sub, filename = 'ep_1x01.en.srt')

        #tmp1.display_graph()

        collection.add_object(epfile, recurse = Equal.OnUnique)

        self.assertEqual(len(collection.find_all(Episode)), 1)
        self.assertEqual(len(collection.find_all(Subtitle)), 1)
        self.assertEqual(len(collection.find_all(Media)), 2)

        print 'collec 2 + epfile'
        #collection.display_graph()

        collection.add_object(subfile, recurse = Equal.OnUnique)

        self.assertEqual(len(collection.find_all(Episode)), 1)
        self.assertEqual(len(collection.find_all(Subtitle)), 1)
        self.assertEqual(len(collection.find_all(Media)), 2)

        print 'collec 2 + subfile'
        #collection.display_graph()

        print 'collec 2 + ep md'
        collection.add_object(ep, recurse = Equal.OnUnique)
        #collection.display_graph()
        self.assertEqual(len(collection.find_all(Episode)), 1)
        self.assertEqual(len(collection.find_all(Subtitle)), 1)
        self.assertEqual(len(collection.find_all(Media)), 2)

        print 'collec 2 + sub md'
        collection.add_object(sub, recurse = Equal.OnUnique)
        #collection.display_graph()
        self.assertEqual(len(collection.find_all(Episode)), 1)
        self.assertEqual(len(collection.find_all(Subtitle)), 1)
        self.assertEqual(len(collection.find_all(Media)), 2)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyGoo - An Object-Graph mapper
# Copyright (c) 2010 Nicolas Wack <wackou@gmail.com>
#
# PyGoo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyGoo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from pygoo.abstractnode import AbstractNode
from pygoo.objectnode import ObjectNode
from pygoo.utils import tolist, toresult, is_literal, to_iterator
from pygoo import ontology
import logging

log = logging.getLogger('pygoo.MemoryObjectNode')

class MemoryObjectNode(ObjectNode):

    def __init__(self, graph, props = list(), _classes = set()):
        # NB: this should go before super().__init__() because we need
        #     self._props and self._classes to exist before we can set
        #     attributes
        self._props = {}
        self._classes = set(_classes)
        super(MemoryObjectNode, self).__init__(graph, props)

        log.debug('MemoryNode.__init__: classes = %s' % list(self._classes))
        graph._nodes.add(self)


    def __eq__(self, other):
        return self is other

    def __hash__(self):
        # TODO: verify me
        return id(self)

    def __setattr__(self, name, value):
        if name in [ '_props', '_classes' ]:
            object.__setattr__(self, name, value)
        else:
            super(MemoryObjectNode, self).__setattr__(name, value)


    ### Ontology methods

    def add_class(self, cls):
        self._classes.add(cls)

    def remove_class(self, cls):
        self._classes.remove(cls)

    def clear_classes(self):
        self._classes = set()

    def classes(self):
        return self._classes

    def isinstance(self, cls):
        return cls in self._classes



    ### Accessing literal properties

    def get_literal(self, name):
        # if name is not a literal, we need to throw an exception
        result = self._props[name]
        if is_literal(result):
            return result
        raise AttributeError

    def set_literal(self, name, value):
        self._props[name] = value

    def literal_keys(self):
        return (k for k, v in self._props.items() if is_literal(v))

    def literal_values(self):
        return (v for v in self._props.values() if is_literal(v))

    def literal_items(self):
        return ((k, v) for k, v in self._props.items() if is_literal(v))



    ### Accessing edge properties

    def add_directed_edge(self, name, other_node):
        # otherNode should always be a valid node
        node_list = tolist(self._props.get(name))
        node_list.append(other_node)
        self._props[name] = toresult(node_list)

    def remove_directed_edge(self, name, other_node):
        # other_node should always be a valid node
        node_list = tolist(self._props.get(name))
        node_list.remove(other_node)
        self._props[name] = toresult(node_list)

        # TODO: we should have this, right?
        if self._props[name] is None:
            del self._props[name]


    def outgoing_edge_endpoints(self, name = None):
        if name is None:
            return self._all_outgoing_edge_endpoints()
        else:
            return self._outgoing_edge_endpoints(name)

    def _outgoing_edge_endpoints(self, name):
        # if name is not an edge, we need to throw an exception
        result = self._props[name]
        if not is_literal(result):
            return to_iterator(result)
        raise AttributeError

    def _all_outgoing_edge_endpoints(self):
        for prop, eps in self.edge_items():
            for ep in eps:
                yield ep


    def edge_keys(self):
        return (k for k, v in self._props.items() if not is_literal(v))

    def edge_values(self):
        return (to_iterator(v) for v in self._props.values() if not is_literal(v))

    def edge_items(self):
        return ((k, to_iterator(v)) for k, v in self._props.items() if not is_literal(v))


    # The next methods are overriden for efficiency
    # They should take precedence over their implementation in ObjectNode, as long as the order
    # of inheritance is respected, ie:
    #
    #   class MemoryObjectNode(MemoryNode, ObjectNode): # GOOD
    #   class MemoryObjectNode(ObjectNode, MemoryNode): # BAD

    def keys(self):
        return self._props.keys()

    # FIXME: wrong implementation as the values for edges should be iterators
    #def values(self):
    #    return self._props.values()

    #def items(self):
    #    return self._props.items()

    def update_valid_classes(self):
        if self.graph()._dynamic:
            self._classes = set(cls for cls in ontology._classes.values() if self.is_valid_instance(cls))
        else:
            # no need to do anything
            pass




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyGoo - An Object-Graph mapper
# Copyright (c) 2010 Nicolas Wack <wackou@gmail.com>
#
# PyGoo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyGoo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from pygootest import *

class TestObjectNode(TestCase):

    def setUp(self):
        ontology.clear()

    def testBasicObjectNode(self):
        g = MemoryObjectGraph()
        n1 = g.BaseObject()
        self.assertEqual(n1.keys(), [])

        n1.title = u'abc'
        self.assertEqual(n1.title, u'abc')

        n2 = g.BaseObject(title = u'abc')
        self.assertEqual(n2.title, u'abc')
        self.assertEqual(n1, n2)

        n1.sameas = n2
        # make sure there's no ambiguity between ObjectNodes and BaseObjects
        self.assertEqual(n1.sameas, n2)
        self.assert_(n2.node in n1.node.sameas)
        self.assert_(n2 not in n1.node.sameas)
        self.assertNotEqual(n1.sameas, n2.node)
        self.assertEqual(BaseObject(n1.sameas), n2)

        n2.plot = u'empty'
        self.assertEqual(n1.sameas.title, u'abc')
        self.assertEqual(n1.sameas.plot, u'empty')

    def testBasicOntology(self):
        class A(BaseObject):
            schema = { 'title': unicode }
            valid = schema.keys()
            unique = valid

        class B(A):
            schema = {}
            #valid = [ 'title' ]
            valid = A.valid
            unique = [ 'title' ]

        class C(BaseObject):
            schema = { 'c': int }
            valid = schema.keys()
            unique = valid

        class D(BaseObject):
            schema = {}
            valid = []

        class E:
            schema = { 'e': int }
            unique = [ 'e' ]

        class F(BaseObject):
            schema = { 'friend': BaseObject }
            reverse_lookup = { 'friend': 'friend' }
            valid = schema.keys()

        self.assertEqual(issubclass(A, BaseObject), True)
        self.assertEqual(issubclass(B, A), True)
        self.assertEqual(issubclass(A, A), True)
        self.assertEqual(issubclass(A, B), False)
        self.assertEqual(issubclass(B, BaseObject), True)
        self.assertEqual(issubclass(C, A), False)
        self.assertEqual(B.parent_class().class_name(), 'A')

        #self.assertRaises(TypeError, ontology.register, E) # should inherit from BaseObject
        # should not define the same reverse_lookup name when other class is a superclass (or subclass) of our class
        self.assertRaises(TypeError, ontology.register, F)

        self.assert_(ontology.get_class('A') is A)
        #self.assertRaises(ValueError, ontology.get_class, 'D') # not registered
        self.assert_(ontology.get_class('B').parent_class().parent_class() is BaseObject)

        # test instance creation
        g = MemoryObjectGraph()
        a = g.A(title = u'Scrubs', epnum = 5)

        self.assertEqual(type(a), A)
        self.assertEqual(a.__class__, A)
        self.assertEqual(a.__class__.class_name(), 'A')
        self.assertEqual(a.__class__.__name__, 'A')

    def registerMediaOntology(self):
        class Series(BaseObject):
            schema = { 'title': unicode,
                       'rating': float
                       }

            valid = [ 'title' ]
            unique = valid

        class Episode(BaseObject):
            schema = { 'series': Series,
                       'season': int,
                       'episodeNumber': int,
                       'title': unicode,
                       'synopsis': unicode
                       }

            reverse_lookup = { 'series': 'episodes' }
            valid = [ 'series', 'season', 'episodeNumber' ]
            unique = valid

            def __cmp__(self, other):
                return (self.season - other.season) or (self.episodeNumber - other.episodeNumber)

        class Person(BaseObject):
            schema = { 'firstName': unicode,
                       'lastName': unicode,
                       'dateOfBirth': float, # TODO: should be datetime
                       }
            valid = [ 'firstName', 'lastName' ]

            def name(self):
                return self.firstName + ' ' + self.lastName

            def best_movie(self):
                """This returns the movie with the highest rating this actor has played in"""
                ratings = [ (role.movie.imdbRating, role.movie) for role in tolist(self.role) ]
                return sorted(ratings)[-1][1]


        class Movie(BaseObject):
            schema = { 'title': unicode,
                       'year': int,
                       'plot': unicode,
                       'imdbRating': float,
                       'director': Person,
                       }
            reverse_lookup = { 'director': 'filmography' }
            valid = [ 'title', 'year' ]
            unique = valid

        class Character(BaseObject):
            schema = { 'name': unicode }
            valid = schema.keys()

        class Role(BaseObject):
            schema = { 'movie': Movie,
                       'actor': Person,
                       'character': Character
                       }
            reverse_lookup = { 'movie': 'roles',
                               'actor': 'actingRoles',
                               'character': 'roles'
                               }
            valid = schema.keys()
            unique = valid


    def testMediaOntology(self):
        self.registerMediaOntology()
        ontology.import_classes([ 'Series', 'Episode', 'Person' ])

        self.assert_('episodes' in Series.schema)
        self.assert_('episodes' in Series.reverse_lookup)
        self.assertEqual(Person.schema['filmography'], ontology.get_class('Movie'))
        self.assertEqual(Person.reverse_lookup['filmography'], 'director')

        g = MemoryObjectGraph(dynamic = True)
        self.createData(g)

        ep = g.find_one(Episode)
        # Series needs only 'title', so it should be a fit if the graph is dynamic
        self.assertEqual(ep.node, Series(ep).node)

        s = g.find_one(Series, title = 'The Wire')
        self.assertRaises(TypeError, Episode, s)


        # test with a static graph
        g = MemoryObjectGraph(dynamic = False)
        self.createData(g)

        ep = g.find_one(Episode)
        # Series needs only 'title', so it should be a fit if the graph is dynamic
        self.assertRaises(TypeError, Series, ep)

        s = g.find_one(Series)
        self.assertRaises(TypeError, Episode, s)


    def testValidUniqueMethods(self):
        self.registerMediaOntology()

        g = MemoryObjectGraph()
        ep = g.Episode(series = g.Series(title = u'House M.D.'), season = 3, episodeNumber = 2)
        self.assertRaises(TypeError, g.Episode, series = u'House M.D.', episodeNumber = 2)
        #self.assertEqual(ep.isValid(), True) # construction of object should fail if not
        ep.title = u'gloub'
        ep.gulp = u'gloubiboulga'
        #self.assertEqual(ep.isValid(), True)
        self.assertRaises(TypeError, setattr, ep, 'title', 3)
        #self.assertEqual(ep.isValid(), False)

        self.assertEqual(ep.is_unique(), True)
        #self.assertEqual(g.Episode(series=u'abc').isUnique(), False)

    def testBasicGraphBehavior(self):
        g = MemoryObjectGraph()
        g2 = MemoryObjectGraph()

        n1 = g.BaseObject(x = 1)
        self.assert_(n1.node in g)
        n2 = g.BaseObject(x = 1)
        n3 = g.BaseObject(n1)
        n4 = g2.BaseObject(n1)

        # equality of objects should be based on value
        self.assertEqual(n1, n2)
        self.assertEqual(n1, n3)
        self.assertEqual(n1, n4)

        # equality of nodes should be based on identity
        self.assertEqual(n1.node, n3.node)
        self.assertNotEqual(n1.node, n2.node)
        self.assertNotEqual(n1.node, n4.node) # not equal as they live in different graphs

        # graph belonging should be tested with identity (related to node)
        self.assert_(n1.node in g)
        self.assert_(n3 in g)
        self.assert_(n4 not in g)


    def testGraphTransfer(self):
        class NiceGuy(BaseObject):
            schema = { 'friend': BaseObject }
            valid = [ 'friend' ]
            reverse_lookup = { 'friend': 'friendOf' }

        g1 = MemoryObjectGraph()
        g2 = MemoryObjectGraph()

        n1 = g1.BaseObject(a = 23)
        n2 = g1.NiceGuy(friend = n1)

        # by default we use recurse = OnIdentity
        # we could also have g2.add_object(n2, recurse = OnValue) or recurse = OnUnique
        r2 = g2.add_object(n2)
        self.assert_(r2 in g2)
        # verify it also brought its friend
        self.assert_(r2.friend in g2)
        self.assertEqual(r2.friend.a, 23)
        self.assertEquals(len(g2.find_all(a = 23)), 1)

        # FIXME: the following fails, it surely hides a bug...
        # if we keep on adding by identity, we will end up with lots of friends with a=23
        #n3 = g1.BaseObject(name = u'other node', friend = n1)
        #r3 = g2.add_object(n3)
        #self.assertEquals(len(g2.find_all(a = 23)), 2)

        # if we keep on adding by identity, we will end up with lots of friends with a=23
        n3 = g1.NiceGuy(name = u'other node', friend = n1)
        r3 = g2.add_object(n3)
        self.assertEquals(len(g2.find_all(a = 23)), 2)

        # if we add and recurse on value, we shouldn't be adding the same node again and again
        n4 = g1.NiceGuy(name = u'3rd of its kind', friend = g1.BaseObject(a = 23))

        r4 = g2.add_object(n4, recurse = Equal.OnValue)

        self.assertEquals(len(g2.find_all(a = 23)), 2) # no new node added with a = 23
        # reference should have been updated though, no trying to keep old friends
        self.assert_(r4.friend.node in [ r.node for r in tolist(r2.friend) ] or
                     r4.friend.node in [ r.node for r in tolist(r3.friend) ])


    def testReverseAttributeLookup(self):
        self.registerMediaOntology()
        ontology.import_classes([ 'Series', 'Episode' ])

        g = MemoryObjectGraph()
        self.createData(g)

        s = g.find_one(Series, title = u'The Wire')
        self.assertEqual(s, s.episodes[0].series)

        ep = g.find_one(Episode)
        self.assert_(ep in ep.series.episodes)

        # make sure auto-reverse name work correctly
        vador = g.BaseObject(side = u'dark')
        luke = g.BaseObject(side = u'light', father = vador)

        self.assertEqual(luke.father.side, u'dark')
        self.assertEqual(vador.isFatherOf.side, u'light')


    def createData(self, g):
        g.Movie(title = u'Fear and Loathing in Las Vegas', year = 1998)
        g.Movie(title = u'The Dark Knight', year = 2008)

        wire = g.Series(title = u'The Wire')

        g.Episode(series = wire,
                  season = 2,
                  episodeNumber = 1,
                  title = u'Ebb Tide')

        g.Episode(series = wire,
                  season = 2,
                  episodeNumber = 2,
                  title = u'Collateral Damage')


    def testFindObjectsInGraph(self):
        self.registerMediaOntology()
        ontology.import_classes([ 'Movie', 'Series', 'Episode', 'Person', 'Character' ])

        g = MemoryObjectGraph()
        self.createData(g)

        self.assertEqual(len(g.find_all(node_type = Movie)), 2)
        self.assertEqual(len(g.find_all(Episode, lambda x: x.season == 2)), 2)
        self.assertEqual(len(g.find_all(Episode, season = 2)), 2)
        self.assertEqual(g.find_one(Episode, season = 2, episodeNumber = 1).title, 'Ebb Tide')
        e = g.find_one(Episode, season = 2, episodeNumber = 1)

        recentMovies = g.find_all(Movie, lambda m: m.year > 2000)
        self.assertEqual(len(recentMovies), 1)
        self.assertEqual(recentMovies[0].title, 'The Dark Knight')

        self.assertEqual(len(g.find_all(Episode, series_title = 'The Wire')), 2)
        thewire = g.find_one(Series, title = 'The Wire')
        self.assertEqual(len(g.find_all(Episode, series = thewire)), 2)
        '''
        g.find_all(Person, role_movie_title = 'The Dark Knight') # role == Role.isPersonOf
        g.find_all(Character, isCharacterOf_movie_title = 'Fear and Loathing in Las Vegas', regexp = True)
        g.find_all(Character, roles_movie_title = 'Fear and loathing.*', regexp = True)
        print c.isCharacterOf.movie.title
        print c.is_character_of.movie.title
        '''

    def testComplexGraph(self):
        self.registerMediaOntology()
        ontology.import_classes([ 'Movie', 'Series', 'Episode', 'Person', 'Character' ])

        g = MemoryObjectGraph()

        e = g.Episode(series = g.Series(title = u'abc'), season = 1, episodeNumber = 23)
        g.find_or_create(Series, title = u'def')

        self.assertEqual(len(g.find_all(Series)), 2)

        e2 = g.Episode(series = g.find_or_create(Series, title = u'abc'), season = 1, episodeNumber = 34)

        s = g.find_one(Series, title = u'abc')
        self.assertEqual(len(s.episodes), 2)

        # test that methos on BaseObjects subclasses are correctly called. In our case: Episode.__cmp__
        ep1 = g.Episode(series = s, season = 1, episodeNumber = 1)
        ep2 = g.Episode(series = s, season = 1, episodeNumber = 2)
        ep3 = g.Episode(series = s, season = 1, episodeNumber = 3)
        ep4 = g.Episode(series = s, season = 1, episodeNumber = 4)
        ep5 = g.Episode(series = s, season = 1, episodeNumber = 5)

        self.assertEqual(sorted(s.episodes), [ ep1, ep2, ep3, ep4, ep5, e, e2 ])


    def atestNeo4j(self):
        from neo4jobjectgraph import Neo4jObjectGraph

        self.registerMediaOntology()
        g = Neo4jObjectGraph('/tmp/gloub')
        g.clear()


        Movie = ontology.get_class('Movie')
        print 'creating movie'
        m = g.Movie(title = u'hello')
        print 'movie created'

        g.close()

        g2 = Neo4jObjectGraph('/tmp/gloub')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyGoo - An Object-Graph mapper
# Copyright (c) 2010 Nicolas Wack <wackou@gmail.com>
#
# PyGoo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyGoo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from pygoo.abstractnode import AbstractNode
from pygoo import ontology
import types

def to_utf8(obj):
    """Converts all unicode strings in the given object to utf-8 strings."""
    if isinstance(obj, unicode):
        return obj.encode('utf-8')
    elif isinstance(obj, list):
        return [ to_utf8(elem) for elem in obj ]
    elif isinstance(obj, dict):
        result = {}
        for key, value in obj.items():
            result[to_utf8(key)] = to_utf8(value)
        return result
    else:
        return obj

def tolist(obj):
    """Return object as a list:
     - if object is None, return the empty list
     - if object is a single object (ie: not a list), return a list with a
       single element being the given object
     - otherwise, (ie: it is a list), return the object itself
    """
    if obj is None:
        return []
    elif isinstance(obj, list):
        return obj
    else:
        return [ obj ]


def toresult(lst):
    """Take a list and return a value depending on the number of elements in
    that list:
     - 0 elements -> return None
     - 1 element  -> return the single element
     - 2 or more elements -> return the original list."""
    if not lst:
        return None
    elif len(lst) == 1:
        return lst[0]
    else:
        return lst


def multi_is_instance(value, cls):
    """Return whether given object is an instance of the given class, of is a
    sequence that contains only objects of the given class."""
    # FIXME: we might be calling this function a little bit too much...
    if isinstance(value, list):
        return all(isinstance(v, cls) for v in value)
    elif isinstance(value, types.GeneratorType):
        # we can't touch the generator otherwise the values will be lost
        return issubclass(cls, AbstractNode) or issubclass(cls, BaseObject) # NB: this behaviour is debatable
    else:
        return isinstance(value, cls)




def to_iterator(obj):
    """Return an iterator over all objects contained in obj, or just over obj
    if it is not a sequence."""
    for i in tolist(obj):
        yield i

def is_of(name):
    """Return the "is-of" name for the given property."""
    return 'is%sOf' % (name[0].upper() + name[1:])

def is_literal(value):
    return (type(value) in ontology.validLiteralTypes or
            value is None or # TODO: is None could be checked here for validity in the schema
            any(multi_is_instance(value, cls) for cls in ontology.validLiteralTypes))


def check_class(name, value, schema, converters = dict()):
    """This function also converts BaseObjects to nodes after having checked
    their class."""

    # always try to autoconvert a string to a unicode
    if isinstance(value, str):
        value = value.decode('utf-8')
    elif multi_is_instance(value, str):
        value = [ v.decode('utf-8') for v in value ]


    def tonodes(v):
        ontology.import_class('BaseObject')
        if isinstance(v, BaseObject):
            return v.node
        elif isinstance(v, list) and v != [] and isinstance(v[0], BaseObject):
            return (n.node for n in v)
        else:
            return v

    if name not in schema or multi_is_instance(value, schema[name]):
        return tonodes(value)

    # try to autoconvert a string to int or float
    if isinstance(value, basestring) and schema[name] in [ int, float ]:
        try:
            return schema[name](value)
        except ValueError:
            pass


    # TODO: use specified converters, when available

    raise TypeError("The '%s' attribute is of type '%s' but you tried to assign it a '%s'"
                    % (name, schema[name], type(value)))


def reverse_lookup(d, cls):
    """Returns a list of tuples used mostly for node creation, where BaseObjects have been replaced with their nodes.
    They are triples of (name, literal value or node generator, reverseName).
    This also checks for type validity and converts the values if they have type converters.
    string -> unicode, string -> int  and  string -> float  are done automatically."""
    return [ (name,
              check_class(name, value, cls.schema, cls.converters),
              cls.reverse_lookup.get(name) or is_of(name))
             for name, value in d.items() ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyGoo - An Object-Graph mapper
# Copyright (c) 2010 Nicolas Wack <wackou@gmail.com>
#
# PyGoo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyGoo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from pygootest import *

class TestMemory(TestCase):

    def setUp(self):
        ontology.clear()

    def testGraphGC(self):
        ontology._graphs.clear()

        def f():
            g = MemoryObjectGraph()

        f()
        # as the graph came out of scope, it should be GC'ed
        self.assertEqual(ontology._graphs.items(), [])

        def g():
            g = MemoryObjectGraph()
            o = g.BaseObject(a = 3)

        g()
        # as the graph came out of scope, it should be GC'ed
        self.assertEqual(ontology._graphs.items(), [])

        g = MemoryObjectGraph()
        o1 = g.BaseObject(a = 3)
        o2 = g.BaseObject(b = 2, c = o1)

        self.assertEqual(len(ontology._graphs.items()), 1)

        del g
        self.assertEqual(ontology._graphs.items(), [])




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyGoo - An Object-Graph mapper
# Copyright (c) 2010 Nicolas Wack <wackou@gmail.com>
#
# PyGoo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyGoo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


from pygootest import *


def print_class(cls):
    print 'class: %s' % cls.__name__
    print 'parent: %s' % cls.parent_class().__name__
    print 'schema', cls.schema
    print 'implicit', cls.schema._implicit
    print 'rlookup', cls.reverse_lookup

class TestInheritance(TestCase):

    def setUp(self):
        ontology.clear()

    def testSchema(self):
        class A(BaseObject):
            schema = { 'a': int }
            valid = []

        self.assertEqual(len(A.schema), 1)
        self.assertEqual(A.schema['a'], int)

        class B(A):
            schema = { 'b': unicode }
            valid = []

        self.assertEqual(len(A.schema), 1)
        self.assertEqual(A.schema['a'], int)

        self.assertEqual(len(B.schema), 2)
        self.assertEqual(B.schema['a'], int)
        self.assertEqual(B.schema['b'], unicode)

        class C(A):
            schema = { 'c': float }
            valid = []

        self.assertEqual(len(A.schema), 1)
        self.assertEqual(A.schema['a'], int)

        self.assertEqual(len(B.schema), 2)
        self.assertEqual(B.schema['a'], int)
        self.assertEqual(B.schema['b'], unicode)

        self.assertEqual(len(C.schema), 2)
        self.assertEqual(C.schema['a'], int)
        self.assertEqual(C.schema['c'], float)

        def multi():
            class D(B, C):
                schema = {}
                valid = []


        self.assertRaises(TypeError, multi)


    def testImplicitSchema(self):
        class A(BaseObject):
            schema = { 'a': int }
            valid = [ 'a' ]

        class E(A):
            schema = {}
            valid = [ 'a' ]

        class B(BaseObject):
            schema = { 'b': float }
            valid = [ 'b' ]

        def testValidVar():
            class B2(B):
                schema = {}
                valid = []

        self.assertRaises(TypeError, testValidVar)

        class C(B):
            schema = { 'friend': A }
            valid = [ 'b' ]
            reverse_lookup = { 'friend': 'friendOf' }

        self.assertEqual(C.schema['friend'], A)
        self.assert_('friend' not in C.schema._implicit)

        class D(A):
            schema = {}
            valid = [ 'a' ]

        self.assertEqual(C.schema['friend'], A)
        self.assert_('friend' not in C.schema._implicit)

        self.assertEqual(D.schema['friendOf'], C)
        self.assertEqual(E.schema['friendOf'], C)
        self.assertEqual(list(D.schema._implicit), ['friendOf'])
        self.assertEqual(list(E.schema._implicit), ['friendOf'])
        self.assertEqual(D.reverse_lookup, { 'friendOf': 'friend' })
        self.assertEqual(E.reverse_lookup, { 'friendOf': 'friend' })


    def testStaticInheritance(self):
        class A(BaseObject):
            schema = { 'a': int }
            valid = []

        class B(BaseObject):
            schema = { 'b': float }
            valid = []

        g = MemoryObjectGraph(dynamic = True)
        b = g.B()
        self.assert_(A in b.node._classes)

        g = MemoryObjectGraph(dynamic = False)
        b = g.B()
        self.assert_(A not in b.node._classes)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyGoo - An Object-Graph mapper
# Copyright (c) 2010 Nicolas Wack <wackou@gmail.com>
#
# PyGoo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyGoo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from pygoo.abstractnode import AbstractNode
from pygoo.utils import is_of, reverse_lookup, toresult, multi_is_instance, check_class
from pygoo import ontology
import types
import logging

log = logging.getLogger('pygoo.BaseObject')


def get_node(node):
    if isinstance(node, AbstractNode):
        return node
    elif isinstance(node, BaseObject):
        return node.node
    elif isinstance(node, list) and isinstance(node[0], BaseObject):
        return [ n.node for n in node ]
    else:
        raise TypeError("Given object is not an ObjectNode or BaseObject instance")

def to_nodes(d):
    result = dict(d)
    for k, v in d.items():
        if isinstance(v, BaseObject):
            result[k] = v.node
        elif isinstance(v, list) and (v != [] and isinstance(v[0], BaseObject)):
            result[k] = [ n.node for n in v ]
    return result


# Metaclass to be used for BaseObject so that they are automatically registered in the ontology
class OntologyClass(type):
    def __new__(cls, name, bases, attrs):
        log.info('Creating class: %s' % name)
        if len(bases) > 1:
            raise TypeError('BaseObject does not allow multiple inheritance for class: %s' % name)
        return type.__new__(cls, name, bases, attrs)

    def __init__(cls, name, bases, attrs):
        log.info('Initializing class: %s' % name)
        super(OntologyClass, cls).__init__(name, bases, attrs)
        ontology.register(cls, attrs)


    def class_variables(cls):
        # need to return a copy here (we're already messing enough with all those mutable objects around...)
        return (ontology.Schema(cls.schema),
                dict(cls.reverse_lookup),
                set(cls.valid),
                set(cls.unique),
                list(cls.order),
                dict(cls.converters))

    def set_class_variables(cls, vars):
        cls.schema = ontology.Schema(vars[0])
        cls.reverse_lookup = dict(vars[1])
        cls.valid = set(vars[2])
        cls.unique = set(vars[3])
        cls.order = set(vars[4])
        cls.converters = set(vars[5])


class BaseObject(object):
    """A BaseObject is a statically-typed object which gets its data from an ObjectNode. In that sense, it
    acts like a view of an ObjectNode, and all data assigned to a BaseObject is actually stored in the
    ObjectNode.

    It is possible and inexpensive to create any number of possibly different BaseObject "views" on an
    ObjectNode, which means that you can dynamically interpret an ObjectNode as being an instance of a
    given class (given that it matches the class schema).

    Equality comparison is done by comparing the properties of the nodes, not the identity of the nodes,
    so two BaseObjects wrapping two different nodes in the graph would be equal if their properties are.

    You should derive from this class to define an ontology domain on ObjectGraphs and ObjectNodes.

    You should define the following class variables in derived classes:

    1- 'schema' which is a dict of property names to their respective types
        ex: schema = { 'epNumber': int,
                       'title': unicode
                       }

    2- 'reverseLookup' which is a dict used to indicate the name to be used for the property name
                       when following a relationship between objects in the other direction.
                       ie: if Episode(...).series == Series('Scrubs'), then we define automatically
                       a way to access the Episode() from the pointed to Series() object.
                       with { 'series': 'episodes' }, we then have:
                       Series('Scrubs').episodes = [ Episode(...), Episode(...) ]
                       reverseLookup must be defined for each property which is a Node object

    3- 'valid' which is the list of properties a node needs to have to be able to be considered
               as a valid instance of this class

    4- 'unique' which is the list of properties that form a primary key

    5- 'order' (optional) which is a list of properties you always want to see in front for debug msgs
               by default, this will be set to the 'valid' variable

    6- 'converters' (optional), which is a dictionary from property name to a pair of functions
                    that are able to serialize/deserialize this property to/from a unicode string.


    Apart from having to define the aforementioned class variables, a BaseObject behaves like a python object,
    so that a subclass of BaseObject can also define its own methods and they can be called normally on instances
    of it.
    As an example, you might want to define:

    class Actor:
        def bestMovies(self):
            maxRating = max(movie.rating for movie in self.roles.movie)
            return [ movie for movie in self.roles.movie if movie.rating == maxRating ]


    Note: at the moment, self.roles.movie wouldn't work, so one would have to do so:
    class Actor:
        def bestMovies(self):
            maxRating = max(role.movie.rating for role in self.roles)
            return [ role.movie for role in self.roles if role.movie.rating == maxRating ]

    Which is slightly less intuitive because it forces us to iterate over the roles instead of over the movies.
    """

    __metaclass__ = OntologyClass

    # TODO: remove those variables which definition should be mandatory
    schema = ontology.Schema({})
    reverse_lookup = {}
    valid = []
    unique = []
    order = []
    converters = {}

    # This variable works just like the 'schema' one, except that it only contains properties which have been defined
    # as a result of the reverseLookup names of other classes
    #_implicitSchema = {}

    def __init__(self, basenode = None, graph = None, allow_incomplete = False, **kwargs):
        #log.debug('%s.__init__: basenode = %s, args = %s' % (self.__class__.__name__, basenode, kwargs))
        if graph is None and basenode is None:
            raise ValueError('You need to specify either a graph or a base node when instantiating a %s' % self.__class__.__name__)

        # just to make sure, while developing. This should probably be removed in a stable version
        if (#(graph is not None and not isinstance(graph, ObjectGraph)) or
            (basenode is not None and not (isinstance(basenode, AbstractNode) or
                                           isinstance(basenode, BaseObject)))):
                raise ValueError('Trying to build a BaseObject from a basenode, but you gave a \'%s\': %s' % (type(basenode).__name__, str(basenode)))

        created = False
        if basenode is None:
            # if no basenode is given, we need to create a new node
            self.node = graph.create_node(reverse_lookup(kwargs, self.__class__),
                                          _classes = ontology.parent_classes(self.__class__))
            created = True

        else:
            basenode = get_node(basenode)

            # if basenode is already in this graph, no need to make a copy of it
            # if graph is None, we might just be making a new instance of a node, so it's in the same graph as well
            if graph is None or graph is basenode.graph():
                self.node = basenode
            else:
                if basenode.edge_keys():
                    # we have links, we can't just create the node without adding the dependencies...
                    raise Exception("sorry, can't do that right now...")

                # TODO: we should be able to construct directly from the other node
                self.node = graph.create_node(reverse_lookup(basenode, self.__class__),
                                              _classes = basenode._classes)
                created = True


            # optimization: avoid revalidating the classes all the time when creating a BaseObject from a pre-existing node
            if kwargs:
                self.update(kwargs)

        # if we just created a node and the graph is static, we gave it its valid classes without actually checking...
        # if not a valid instance, remove it from the list of valid classes so that the next check will fail
        if created and not self.node.graph()._dynamic:
            if allow_incomplete and not self.node.has_valid_properties(self.__class__, set(self.__class__.valid).intersection(set(self.node.keys()))):
                self.node.remove_class(self.__class__)
            if not allow_incomplete and not self.node.is_valid_instance(self.__class__):
                self.node.remove_class(self.__class__)


        # make sure that the new instance we're creating is actually a valid one
        # Note: the following comment shouldn't be necessary if the list of valid classes is always up-to-date
        #if not (self.node.isinstance(self.__class__) or
        #        (self.node._graph._dynamic and self.node.isValidInstance(self.__class__))):
        if not self.node.isinstance(self.__class__):
            # if we just created the node and it is invalid, we need to remove it
            if created:
                self.node.graph().delete_node(self.node)

            raise TypeError("Cannot instantiate a valid instance of %s because:\n%s" %
                            (self.__class__.__name__, self.node.invalid_properties(self.__class__)))



    def __contains__(self, prop):
        return prop in self.node

    def get(self, name):
        try:
            return getattr(self, name)
        except:
            return None

    def __getattr__(self, name):
        result = getattr(self.node, name)

        # if the result is an ObjectNode, wrap it with the class it has been given in the class schema
        # if it was not in the class schema, simply returns an instance of BaseObject
        if isinstance(result, types.GeneratorType):
            result_class = self.__class__.schema.get(name) or BaseObject
            return toresult([ result_class(basenode = n) for n in result ])

        # FIXME: better test here (although if the graph is consistent (ie: always returns generators) it shouldn't be necessary)
        #elif isinstance(result, list) and isinstance(result[0], AbstractNode):
        #    resultClass = self.__class__.schema.get(name) or BaseObject
        #    return [ resultClass(basenode = node) for node in result ]

        else:
            return result


    def __setattr__(self, name, value):
        if name == 'node':
            object.__setattr__(self, name, value)
        else:
            self.set(name, value)

    def _apply_multi(self, func, name, value, validate):
        cls = self.__class__
        if name in cls.schema._implicit:
            raise ValueError("Implicit properties are read-only (%s.%s)" % (cls.__name__, name))

        # if mode == STRICT and name not in cls.schema:
        #     raise ValueError("Unknown property in the class schema: (%s.%s)" % (cls.__name__, name))

        # objects are statically-typed here; graphType == 'STATIC'
        # this also converts value to the correct type if an autoconverter was given in the class definition
        # and replaces BaseObjects with their underlying nodes.
        value = check_class(name, value, cls.schema)

        # if we don't have a reverse lookup for this property, default to a reverse name of 'is%(Property)Of'
        reverse_name = self.__class__.reverse_lookup.get(name) or is_of(name)

        func(name, value, reverse_name, validate)

    def set(self, name, value, validate = True):
        """Sets the given value to the named property."""
        self._apply_multi(self.node.set, name, value, validate)

    def append(self, name, value, validate = True):
        self._apply_multi(self.node.append, name, value, validate)

    def __eq__(self, other):
        # TODO: should allow comparing a BaseObject with an ObjectNode?
        if not isinstance(other, BaseObject): return False

        if self.node == other.node: return True

        # FIXME: this could lead to cycles or very long chained __eq__ calling on properties
        return self.explicit_items() == other.explicit_items()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return self.node.to_string(cls = self.__class__, default = BaseObject).encode('utf-8')

    def __repr__(self):
        return self.__str__()

    def schema_keys(self):
        return self.__class__.schema.keys()

    def schema_items(self):
        return [ x for x in self.items() if x[0] in self.__class__.schema ]

    def explicit_schema_keys(self):
        return [ x for x in self.schema_keys() if x not in self.__class__.schema._implicit ]

    def explicit_schema_items(self):
        return [ x for x in self.schema_items() if x[0] not in self.__class__.schema._implicit ]

    def explicit_keys(self):
        return [ x for x in self.keys() if x not in self.__class__.schema._implicit ]

    def explicit_items(self):
        return [ x for x in self.items() if x[0] not in self.__class__.schema._implicit ]

    @classmethod
    def class_name(cls):
        return cls.__name__

    @classmethod
    def parent_class(cls):
        return cls.__mro__[1]

    def virtual(self):
        """Return an instance of the most specialized class that this node is an instance of."""
        return self.node.virtual()

    def update(self, props):
        props = to_nodes(props)
        for name, value in props.items():
            self.set(name, value, validate = False)
        self.node.update_valid_classes()

    def is_unique(self):
        """Return whether all unique properties (as defined by the class) of the ObjectNode
        are non-null."""
        for prop in self.__class__.unique:
            if prop not in self.node:
                return False
        return True

    def unique_properties(self):
        return self.__class__.unique

    def unique_key(self):
        """Return a tuple containing an unique identifier (inside its class) for this instance.
        If some unique fields are not specified, None will be put instead."""
        return tuple(self.get(k) for k in self.__class__.unique)


    def ordered_properties(self):
        """Returns the list of properties ordered using the defined order in the subclass.

        NB: this should be replaced by using an OrderedDict."""
        result = []
        property_names = list(self.node.keys())

        for p in self.__class__.order:
            if p in property_names:
                result.append(p)
                property_names.remove(p)

        return result + property_names



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyGoo - An Object-Graph mapper
# Copyright (c) 2010 Nicolas Wack <wackou@gmail.com>
#
# PyGoo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyGoo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import logging

log = logging.getLogger('pygoo.AbstractDirectedGraph')

# Nice & clean enum implementation by Gabriel Genellina, MIT licensed
# as found at http://code.activestate.com/recipes/577024-yet-another-enum-for-python/
def enum(typename, field_names):
    """Create a new enumeration type"""

    if isinstance(field_names, str):
        field_names = field_names.replace(',', ' ').split()
    d = dict((reversed(nv) for nv in enumerate(field_names)), __slots__ = ())
    return type(typename, (object,), d)()



Equal = enum('Equal',
             'OnIdentity, OnValue, OnValidValue, OnUnique, OnLiterals')


class AbstractDirectedGraph(object):
    """This class describes a basic directed graph, with the addition that the nodes
    have a special property "classes" which describe which class a node can "morph"
    into. This mechanism could be built as another basic node property, but it is
    not for performance reasons.

    The AbstractDirectedGraph class is the basic interface one needs to implement
    for providing a backend storage of the data. It is complementary to the
    AbstractNode interface.

    You only need to provide the implementation for this interface, and then an
    ObjectGraph can be automatically built upon it.

    The methods you need to implement fall into the following categories:
     - create / delete node(s)
     - get all nodes / only nodes from a given class
     - check whether a node lives in a given graph
     - find a node given a list of matching properties
    """

    def clear(self):
        """Delete all nodes and links in this graph."""
        raise NotImplementedError

    def create_node(self, props = list(), _classes = set()):
        """Create a node with the given properties and classes."""
        raise NotImplementedError

    def delete_node(self, node):
        """Remove a given node from this graph.

        Strategies for what to do with linked nodes should be configurable, ie:
        remove incoming/outgoing linked nodes as well, only remove link but do
        not touch linked nodes, etc..."""
        raise NotImplementedError

    def nodes(self):
        """Return an iterable on all the nodes in the graph."""
        raise NotImplementedError

    def nodes_from_class(self, cls):
        """Return an iterable on the nodes of a given class."""
        raise NotImplementedError


    def add_directed_edge(self, node, name, other_node):
        # other_node should always be a valid node
        node.add_directed_edge(name, other_node)

    def remove_directed_edge(self, node, name, other_node):
        # other_node should always be a valid node
        node.remove_directed_edge(name, other_node)

    def contains(self, node):
        """Return whether this graph contains the given node.

        multiple strategies can be used here for determing object equality, such
        as all properties equal, the primary subset of properties equal, etc...
        (those are defined by the ObjectNode)"""
        raise NotImplementedError

    def __contains__(self, node):
        """Return whether this graph contains the given node (identity)."""
        # TODO: remove this, as it should only be implemented in the ObjectGraph
        raise NotImplementedError


    ### Methods related to a graph serialization

    def to_nodes_and_edges(self):
        """Convert a graph in a tuple (nodes, edges, classes).

        This is used for serializing a graph, as the resulting tuple is picklable,
        for instance.
        """
        nodes = {}
        classes = {}
        rnodes = {}
        edges = []

        i = 0
        for n in self.nodes():
            nodes[i] = list(n.literal_items())
            rnodes[id(n)] = i
            classes[i] = [ cls.__name__ for cls in n._classes ]
            i += 1

        for n in self.nodes():
            for prop, links in n.edge_items():
                for other_node in links:
                    edges.append((rnodes[id(n)], prop, rnodes[id(other_node)]))

        return nodes, edges, classes

    def from_nodes_and_edges(self, nodes, edges, classes):
        """Create a graph from the given (nodes, edges, classes) tuple."""
        from pygoo import ontology

        self.clear()
        idmap = {}
        for _id, node in nodes.items():
            idmap[_id] = self.create_node(props = ((prop, value, None) for prop, value in node),
                                          _classes = (ontology.get_class(cls) for cls in classes[_id]))

        for node, name, other_node in edges:
            idmap[node].add_directed_edge(name, idmap[other_node])

        # TODO: we need to make sure that the current ontology is the saem as when we saved this graph, otherwise
        #       previously set classes might not be valid anymore, or some subclasses won't be correctly set
        # we need to revalidate explicitly, as we might have classes in our ontolo
        #self.revalidate_objects()


    def save(self, filename):
        """Saves the graph to the given filename."""
        import cPickle as pickle
        pickle.dump(self.to_nodes_and_edges(), open(filename, 'w'))

    def load(self, filename):
        import cPickle as pickle
        self.from_nodes_and_edges(*pickle.load(open(filename)))

    # __getstate__ and __setstate__ are needed for the cache to be able to work
    def __getstate__(self):
        return self.to_nodes_and_edges()

    def __setstate__(self, state):
        # FIXME: this doesn't belong here...
        self._dynamic = False

        self.from_nodes_and_edges(*state)


    ### Utility methods

    def display_graph(self, title = ''):
        import cPickle as pickle
        import tempfile
        import subprocess

        nodes, edges, classes = self.to_nodes_and_edges()
        fid, filename = tempfile.mkstemp(suffix = '.png')

        dg = []
        dg += [ 'digraph G {' ]
        dg += [ '  label="%s"' % title ]

        def tostring(prop):
            if isinstance(prop, list):
                result = u', '.join([ unicode(p) for p in prop ])
            else:
                result = unicode(prop)

            # appply html chars transformation
            result = result.replace('&', '&amp;')

            return result

        for _id, n in nodes.items():
            label = '<FONT COLOR="#884444">%s</FONT><BR/>' % (', '.join(cls for cls in classes[_id] if cls != 'BaseObject') or 'BaseObject')
            label += '<BR/>'.join([ '%s: %s' % (name, tostring(prop)[:32].encode('utf-8')) for name, prop in n ])
            dg += [ 'node_%d [shape=polygon,sides=4,label=<%s>];' % (_id, label) ]

        for node, name, other_node in edges:
            dg += [ 'node_%d -> node_%d [label="%s"];' % (node, other_node, name) ]

        dg += [ '}' ]

        subprocess.Popen([ 'dot', '-Tpng', '-o', filename ], stdin = subprocess.PIPE).communicate('\n'.join(dg))

        if sys.platform == 'linux2':
            subprocess.Popen([ 'gwenview', filename ], stdout = subprocess.PIPE, stderr = subprocess.PIPE).communicate()
        else:
            subprocess.Popen([ 'open', filename ], stdout = subprocess.PIPE, stderr = subprocess.PIPE).communicate()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from neo4j import NeoService
import subprocess, glob
import logging

log = logging.getLogger('pygoo.Neo4j')


dirname = None
graph = None
transaction = None


def default_jvm():
    # make an educated guess on debian-based system
    java = subprocess.Popen([ 'ls', '-l', '/etc/alternatives/java' ], stdout = subprocess.PIPE).stdout.read().strip().split()[-1]
    java_rootdir = '/'.join(java.split('/')[:-2])
    jvms = glob.glob(java_rootdir + '/lib/*/client/libjvm.so')

    if jvms:
        return jvms[0]

    raise OSError, 'Could not find default JVM'



def close():
    global dirname, graph, transaction

    try:
        graph.shutdown()
        dirname, graph, transaction = None, None, None
    except RuntimeError, e:
        log.error(str(e))


def open(dbpath):
    global dirname, graph, transaction

    if dirname:
        raise RuntimeError("A Neo4j database is already open. Please close it before opening a new one.")

    dirname = dbpath
    graph = NeoService(dbpath, jvm = default_jvm())
    transaction = graph.transaction


def delete_all_data():
    global dirname
    dbpath = dirname

    close()

    print dbpath, type(dbpath)
    # do it the brute-force way...
    subprocess.call([ 'rm', '-fr', dbpath ])

    open(dbpath)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyGoo - An Object-Graph mapper
# Copyright (c) 2010 Nicolas Wack <wackou@gmail.com>
#
# PyGoo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyGoo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from pygootest import *

class TestAbstractNode(TestCase):

    def setUp(self):
        ontology.clear()

    def testMRO(self):
        class A(object):
            def __init__(self):
                print 'A.__init__()'

            def __contains__(self, obj):
                print 'A.__contains__'

        class B(A):
            def __init__(self):
                super(B, self).__init__()
                print 'B.__init__()'

            #def __contains__(self, obj):
            #    print 'B.__contains__'

        class C(A):
            def __init__(self):
                super(C, self).__init__()
                print 'C.__init__()'

            def __contains__(self, obj):
                print 'C.__contains__'

        class D(B, C):
            def __init__(self):
                super(D, self).__init__()
                print 'D.__init__()'

            #def __contains__(self, obj):
            #    print 'D.__contains__'

        d = D()
        3 in d

    def testAbstractNode(self, GraphClass = MemoryObjectGraph):
        g = GraphClass()

        n = g.create_node()
        n.set_literal('title', u'abc')
        self.assertEqual(n.get_literal('title'), u'abc')
        self.assertEqual(list(n.literal_keys()), [ 'title' ])
        self.assertEqual(list(n.edge_keys()), [])

        n2 = g.create_node()
        n.add_directed_edge('friend', n2)
        n2.add_directed_edge('friend', n)
        self.assertEqual(n.get_literal('title'), u'abc')
        self.assertEqual(list(n.literal_keys()), [ 'title' ])
        self.assertEqual(list(n.edge_keys()), [ 'friend' ])
        self.assertEqual(list(n.outgoing_edge_endpoints('friend')), [ n2 ])
        self.assertEqual(list(n2.outgoing_edge_endpoints('friend')), [ n ])

        n3 = g.create_node()
        n.add_directed_edge('friend', n3)
        n3.add_directed_edge('friend', n)
        self.assertEqual(len(list(n.outgoing_edge_endpoints('friend'))), 2)
        self.assertEqual(len(list(n.outgoing_edge_endpoints())), 2)
        self.assert_(n2 in n.outgoing_edge_endpoints('friend'))
        self.assert_(n3 in n.outgoing_edge_endpoints('friend'))


    def testBasicObjectNode(self, ObjectGraphClass = MemoryObjectGraph):
        g = ObjectGraphClass()

        n = g.create_node()
        n.title = u'abc'
        self.assertEqual(n.get_literal('title'), u'abc')
        self.assertEqual(list(n.literal_keys()), [ 'title' ])
        self.assertEqual(list(n.edge_keys()), [])

        n2 = g.create_node()
        n.friend = n2
        self.assertEqual(n.title, u'abc')
        self.assertEqual(list(n.literal_keys()), [ 'title' ])
        self.assertEqual(list(n.edge_keys()), [ 'friend' ])
        self.assertEqual(list(n.friend), [ n2 ])
        self.assertEqual(list(n2.isFriendOf), [ n ])

        n3 = g.create_node()
        n.friend = [ n2, n3 ]
        self.assert_(n in n2.isFriendOf)
        self.assert_(n in n3.isFriendOf)
        self.assertEqual(tolist(n3.get('friend')), [])

        n4 = g.create_node()
        n4.friend = n.friend
        self.assert_(n in n2.isFriendOf)
        self.assert_(n in n3.isFriendOf)
        self.assert_(n4 in n2.isFriendOf)
        self.assert_(n4 in n3.isFriendOf)

        n.friend = []
        self.assert_(n not in n2.isFriendOf)
        self.assert_(n not in n3.isFriendOf)
        self.assert_(n4 in n2.isFriendOf)
        self.assert_(n4 in n3.isFriendOf)


    def testBaseObject(self, GraphClass = MemoryObjectGraph):
        class NiceGuy(BaseObject):
            schema = { 'friend': BaseObject }
            valid = [ 'friend' ]
            reverse_lookup = { 'friend': 'friendOf' }

        # There is a problem when the reverse-lookup has the same name as the property because of the types:
        # NiceGuy.friend = BaseObject, BaseObject.friend = NiceGuy
        #
        # it should also be possible to have A.friend = B and C.friend = B, and not be a problem for B, ie: type(B.friend) in [ A, C ]
        #
        # or we should restrict the ontology only to accept:
        #  - no reverseLookup where key == value
        #  - no 2 classes with the same link types to a third class
        # actually, no reverseLookup where the implicit property could override an already existing one

        g1 = GraphClass()
        g2 = GraphClass()

        n1 = g1.BaseObject(n = u'n1', a = 23)
        n2 = g1.NiceGuy(n = u'n2', friend = n1)
        self.assertEqual(n1.friendOf, n2)

        r2 = g2.add_object(n2)
        r2.n = u'r2'
        self.assertEqual(n1.friendOf, n2)

        n3 = g1.NiceGuy(name = u'other node', friend = n1)
        r3 = g2.add_object(n3)

        # TODO: also try adding n2 after n3 is created

        o1 = g1.BaseObject(n = u'o1')
        o2 = g1.BaseObject(n = u'o2')

        old = n3.friend
        n3.friend = [ o1, o2 ]
        self.assertEqual(o1.friendOf, n3)
        self.assertEqual(o2.friendOf, n3)
        self.assertEqual(tolist(old.friendOf), [n2])
        self.assertEqual(old.friendOf, n2)

        n4 = g1.NiceGuy(n = u'n4', friend = n3.friend)
        self.assert_(o1 in n4.friend)
        self.assert_(o2 in n4.friend)
        self.assert_(n3 in o1.friendOf)
        self.assert_(n3 in o2.friendOf)
        self.assert_(n4 in o1.friendOf)
        self.assert_(n4 in o2.friendOf)

        n3.friend = []
        self.assertEqual(o1.friendOf, n4)
        self.assertEqual(o2.friendOf, n4)

        g1.save('/tmp/pygoo_unittest.db')

        g3 = GraphClass()
        g3.load('/tmp/pygoo_unittest.db')

        self.assertEqual(g3.find_one(NiceGuy, n = 'n2').friend.a, 23)
        self.assertEqual(g3.find_one(NiceGuy, n = 'n2').friend.node, g3.find_one(BaseObject, n = 'n1').node)

        os.remove('/tmp/pygoo_unittest.db')


    def testAddObject(self):
        ontology.import_ontology('media')

        g = MemoryObjectGraph()

        g1 = MemoryObjectGraph()
        m1 = g1.Media(filename = 'a.avi',
                      metadata = g1.Episode(series = g1.Series(title = 'A'),
                                            season = 1,
                                            episodeNumber = 1))
        g.add_object(m1, recurse = Equal.OnUnique)

        g2 = MemoryObjectGraph()
        m2 = g2.Media(filename = 'a.srt',
                      metadata = g2.Subtitle(language = 'en',
                                             metadata = g2.Episode(series = g2.Series(title = 'A'),
                                                                   season = 1,
                                                                   episodeNumber = 1)))

        g.add_object(m2, recurse = Equal.OnUnique)

        self.assertEqual(len(g.find_all(Episode)), 1)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyGoo - An Object-Graph mapper
# Copyright (c) 2010 Nicolas Wack <wackou@gmail.com>
#
# PyGoo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyGoo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

log = logging.getLogger('pygoo.AbstractNode')



class AbstractNode(object):
    """This class describes a basic node in a directed graph, with the addition
    that it has a special property "classes" which describe which class this node
    can "morph" into. This mechanism could be built as another basic node property,
    but it is not for performance reasons.

    It can also have named directed edges to other nodes, ie: the edges are not
    "anonymous", but are first-class citizens (in future versions, we might even
    want to add other properties to an edge).
    A node can have any number of edges of the same type to other nodes (even
    multiples edges to the same node), but it can't have edges to itself.

    The AbstractNode class is the basic interface one needs to implement for
    providing a backend storage of the data. It is complementary to the
    AbstractDirectedGraph interface.

    You only need to provide the implementation for this interface, and then an
    ObjectNode can be automatically built upon it.

    The methods you need to implement fall into the following categories:
     - add / remove / clear / check / get current valid classes
     - get / set / iterate over literal properties
     - add / remove / iterate over edges

    Note: as ObjectNode reimplements the __setattr__ method, you have to do the same
          in your subclass of AbstractNode to catch the instance attributes you need
          to be able to set. Failure to do this will most certainly result in an
          infinite loop, ie: it looks like everything is stuck very soon, or even a
          stack overflow due to infinite recursion.

    """

    def __init__(self, graph, props = list()):
        log.debug('AbstractNode.__init__: graph = %s' % str(graph))


    def __eq__(self, other):
        """Return whether two nodes are equal.

        This should implement identity of nodes, not properties equality (this should
        be done in the BaseObject instance)."""
        raise NotImplementedError

    def __hash__(self):
        raise NotImplementedError

    ## Methods needed for storing the nodes ontology (caching)
    ## Note: this could be implemented only using literal values, but it is left
    ##       as part of the API as this is something which is used a lot and
    ##       benefits a lot from being optimized, which can be more easily done
    ##       in the implementation

    def add_class(self, cls):
        """Add the given class to the list of valid classes for this node."""
        raise NotImplementedError

    def remove_class(self, cls):
        """Remove the given class from the list of valid classes for this node."""
        raise NotImplementedError

    def clear_classes(self):
        """Clear the current list of valid classes."""
        raise NotImplementedError

    def classes(self):
        """Return an iterator over the list of classes."""
        raise NotImplementedError

    def isinstance(self, cls):
        """Return whether this node contains the given class in its list of valid classes."""
        raise NotImplementedError


    ### Methods related to getting / setting literal properties

    def get_literal(self, name):
        """Return the literal with the given name.

        :raises AttributeError: If the given name doesn't correspond to a literal property of this node.
        """
        raise NotImplementedError

    def set_literal(self, name, value):
        """Sets the literal to the given value.
        This method can assume that value is always one of the valid literal types."""
        raise NotImplementedError

    def literal_keys(self):
        """Return an iterable over the literal names."""
        raise NotImplementedError

    def literal_values(self):
        """Return an iterable over the literal values."""
        raise NotImplementedError

    def literal_items(self):
        """Return an iterable over the literal items (key, value)."""
        raise NotImplementedError


    ## Methods related to getting / setting edges (edge properties that point
    ## to other nodes)

    def add_directed_edge(self, name, other_node):
        """Add a named edge from this node to the given one.
        TODO: both nodes should live in the same graph, right? otherwise we throw an exception?"""
        raise NotImplementedError

    def remove_directed_edge(self, name, other_node):
        """Remove the given edge between the 2 nodes.
        TODO: should we throw an exception if it doesn't exist or ignore it?"""
        raise NotImplementedError

    def outgoing_edge_endpoints(self, name = None):
        """Return all the nodes which this node points to with the given edge type.
        If name is None, return all outgoing edge points."""
        # Note: it is *imperative* that this function return a generator and not
        #       just any iterable over the values
        raise NotImplementedError

    def edge_keys(self):
        """Return an iterable over the edge names."""
        raise NotImplementedError

    def edge_values(self):
        """Return an iterable over the edge values (ie: the nodes pointed to by this one)."""
        # Note: it is *imperative* that this function return a generator for each
        #       value and not just any iterable over the values
        raise NotImplementedError

    def edge_items(self):
        """Return an iterable over the edge items (name, nodes pointed to)."""
        # Note: it is *imperative* that this function return a generator for each
        #       value and not just any iterable over the values
        raise NotImplementedError


    ### Additional utility methods

    def unlink_all(self):
        """Remove all incoming and outgoing edges from this node."""
        for name, nodes in self.edge_items():
            for n in nodes:
                self.remove_directed_edge(name, n)
                for oname, onodes in n.edge_items():
                    for n2 in onodes:
                        if n2 == self:
                            n.remove_directed_edge(oname, self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyGoo - An Object-Graph mapper
# Copyright (c) 2010 Nicolas Wack <wackou@gmail.com>
#
# PyGoo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyGoo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import with_statement
from pygoo.memoryobjectnode import MemoryObjectNode
from pygoo.objectnode import ObjectNode
from pygoo.utils import toresult
import neo
import logging

log = logging.getLogger('pygoo.Neo4jObjectNode')


# SYNC can take 2 values:
# - AUTO: automatically synchronize changes with the underlying store as they are modified
# - MANUAL: only flush data to store when the synchronize() method is called
#
# when SYNC == 'AUTO', there is another option to get rid of the calls to the caching MemoryObjectNode (ie: pure database-based)


class Neo4jObjectNode(MemoryObjectNode):
    """This is a proxy class for the neo4j instances.
    It derives from the MemoryObjectNode, which serves as a cache.
    This only loads the nodes needed from the DB when it needs them."""

    def __init__(self, graph, props = [], neonode = None):
        if neonode is not None:
            self._neonode = neonode
            return

        literal = {}
        links = []
        for prop, value, reverse_name in props:
            if isinstance(value, ObjectNode):
                links.append((prop, value, reverse_name))
            else:
                literal[prop] = value

        print 'starting transaction'
        with neo.transaction:
            self._neonode = neo.graph.node(**literal)
            print 'node created'
            self._neonode._ingraph(neo.graph.reference_node) # as a way to get them all
            print 'setting links'
            for link in links:
                self.setLink(*link)
            print 'setting links ok'
        print 'transaction ok'

        #MemoryObjectNode.__init__(self, graph, props)


    def __eq__(self, other):
        return self._neonode == other._neonode

    def __hash__(self):
        # TODO: verify me
        return id(self)

    ### Acessing properties methods


    def __setattr__(self, name, value):
        if name == '_neonode':
            object.__setattr__(self, name, value)
        else:
            MemoryObjectNode.__setattr__(self, name, value)

    def get_link(self, name):
        log.debug('getlink:', name)
        return toresult([ r.end for r in self._neonode.relationships() if r.type == name ])

    def get_literal(self, name):
        log.debug('getliteral:', name)
        result = self._neonode[name]
        if isinstance(result, basestring):
            return result

        return result.value # fix in neo4j.py

    def set_literal(self, name, value):
        MemoryObjectNode.set_literal(self, name, value) # keep in cache
        with neo.transaction:
            self._neonode[name] = value

    #def update(self, props):



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyGoo - An Object-Graph mapper
# Copyright (c) 2010 Nicolas Wack <wackou@gmail.com>
#
# PyGoo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyGoo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from pygoo.memoryobjectgraph import MemoryObjectGraph
from pygoo.neo4jobjectnode import Neo4jObjectNode
import neo4j
import neo
import logging

log = logging.getLogger('pygoo.Neo4jObjectGraph')

class AllNodes(neo4j.Traversal):
    types = [
        neo4j.Incoming._ingraph,
        ]
    order = neo4j.DEPTH_FIRST
    stop = neo4j.STOP_AT_END_OF_GRAPH
    returnable = neo4j.RETURN_ALL_BUT_START_NODE

def all_nodes():
    # FIXME: ATM returns neo nodes, not expected Neo4jObjectNodes
    return AllNodes(neo.graph.reference_node)

class Neo4jObjectGraph(MemoryObjectGraph):
    """A Neo4jObjectGraph is an ObjectGraph where all data is persistent on disk.
    All attribute modifications are immediately synchronized on the data store.

    A Neo4jObjectGraph uses PersistentObjectNodes."""
    _object_node_class = Neo4jObjectNode

    def __init__(self, dbpath):
        MemoryObjectGraph.__init__(self)
        neo.open(dbpath)

    def close():
        neo.close()

    def clear(self):
        """Delete all objects in this graph."""
        MemoryObjectGraph.clear(self)

        neo.delete_all_data()

    def __contains__(self, node):
        """Return whether this graph contains the given node (identity)."""
        # TODO: only works if correctly cached
        return get_node(node) in self._nodes


    def nodes(self):
        for n in all_nodes():
            yield Neo4jObjectNode(neonode = n)


    def remove_directed_edge(self, node, name, other_node):
        MemoryObjectGraph.remove_directed_edge(self, node, name, other_node)

        for r in node._neonode.relationships():
            if r.end == other_node and r.type == name:
                r.delete()
                # return early, which means if we had the same link multiple times,
                # we only remove one of them. This could be used for doing ref-counting if necessary
                return

    def add_directed_edge(self, node, name, other_node):
        MemoryObjectGraph.add_directed_edge(self, node, name, other_node)

        setattr(node._neonode, name, other_node._neonode)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyGoo - An Object-Graph mapper
# Copyright (c) 2010 Nicolas Wack <wackou@gmail.com>
#
# PyGoo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyGoo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from pygoo.objectnode import ObjectNode
from pygoo.abstractdirectedgraph import AbstractDirectedGraph
from pygoo.memoryobjectnode import MemoryObjectNode
from pygoo.objectgraph import ObjectGraph
import logging

log = logging.getLogger('pygoo.MemoryObjectGraph')


class MemoryObjectGraph(ObjectGraph):
    _object_node_class = MemoryObjectNode

    def __init__(self, **kwargs):
        super(MemoryObjectGraph, self).__init__(**kwargs)
        self._nodes = set()

    def clear(self):
        self._nodes.clear()

    def create_node(self, props = [], _classes = set()):
        return self.__class__._object_node_class(self, props, _classes)

    def delete_node(self, node):
        node.unlinkAll()
        node.graph = None
        self._nodes.remove(node)

        # FIXME: we need to revalidate the touched nodes


    def nodes(self):
        for node in self._nodes:
            yield node

    def nodes_from_class(self, cls):
        return (node for node in self._nodes if node.isinstance(cls))

    def contains(self, node):
        """Return whether this graph contains the given node (identity)."""
        return node in self._nodes

    # __getstate__ and __setstate__ are needed for the cache to be able to work
    def __setstate__(self, state):
        self._nodes = set()
        super(MemoryObjectGraph, self).__setstate__(state)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyGoo - An Object-Graph mapper
# Copyright (c) 2010 Nicolas Wack <wackou@gmail.com>
#
# PyGoo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyGoo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import weakref
import logging
import sys

log = logging.getLogger('pygoo.Ontology')

# use dict here for fast text based access (when instantiating objects through a graph, for instance)
_classes = {}
_graphs = weakref.WeakValueDictionary()

# Note: voluntarily omit to put str as allowed types, unicode is much better
#       and it will save us a *lot* of trouble
# Note: list should only be list of literal values
# TODO: add datetime
validLiteralTypes = [ unicode, int, long, float ] #, list ]





class Schema(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self._implicit = set(args[0]._implicit) if args and isinstance(args[0], Schema) else set()

# dict of name to ontologies
# ontologies are (_classes, class variables)
_saved_ontologies = {}

def revalidate_graphs():
    """revalidate all ObjectNodes in all registered Graphs."""
    for g in _graphs.values():
        g.revalidate_objects()

def save_current_ontology(name):
    # we need to make a copy of the elements here, and of the properties explicitly
    # as when we change the ontology, we modify references to always the same class,
    # not a copy of it
    classes = dict(_classes)
    classvars = [ (cls, cls.class_variables()) for cls in _classes.values() ]
    _saved_ontologies[name] = (classes, classvars)

def reload_saved_ontology(name):
    global _classes

    try:
        classes, classvars = _saved_ontologies[name]
    except KeyError:
        raise KeyError("Could not find '%s' ontology" % name)

    _classes = dict(classes)
    for cls, cvars in classvars:
        cls.set_class_variables(cvars)

    revalidate_graphs()

def clear():
    reload_saved_ontology('origin')
    # FIXME: this still leaks memory, as the nodes in a graph have a ref to it
    #_graphs = weakref.WeakValueDictionary()


def subclasses(cls):
    '''Returns the given class and all of its subclasses'''
    return (c for c in _classes.values() if issubclass(c, cls))

def parent_classes(cls):
    '''Returns the given class and all of its parent classes (BaseObject being the topmost class).'''
    return (c for c in _classes.values() if issubclass(cls, c))

def validate_class_definition(cls, attrs):
    BaseObject = _classes['BaseObject']

    valid_types = validLiteralTypes + [ BaseObject ]

    if not issubclass(cls, BaseObject):
        raise TypeError, "'%s' needs to derive from ontology.BaseObject or one of its subclasses" % cls.__name__

    parent = cls.parent_class()

    def check_present(cls, var, ctype, default_value = True):
        if not isinstance(ctype, tuple):
            ctype = (ctype,)

        try:
            value = getattr(cls, var)
        except AttributeError:
            raise TypeError("Your subclass '%s' should define the '%s' class variable as a %s" % (cls.__name__, var, ' or a '.join(c.__name__ for c in ctype)))

        #if not explicitAttribute(cls, var):
        if var not in attrs:
            if default_value:
                # if not explicitly in subclass definition, create a default one
                setattr(cls, var, ctype[0]())
            else:
                raise TypeError("Your subclass '%s' should define explicitly the '%s' class variable as a %s" % (cls.__name__, var, ' or a '.join(c.__name__ for c in ctype)))

        if type(value) not in ctype:
            raise TypeError("Your subclass '%s' defines the '%s' class variable as a '%s', but it should be of type %s" % (cls.__name__, var, type(value).__name__, ' or '.join(c.__name__ for c in ctype)))

        # convert to our preferred type (first of the possible types)
        setattr(cls, var, ctype[0](value))


    def check_schema_subset(cls, var, default_value = True):
        check_present(cls, var, (set, list), default_value)
        for prop in getattr(cls, var):
            if not prop in cls.schema:
                raise TypeError("In '%s': when defining '%s', you used the '%s' variable, which is not defined in the schema" % (cls.__name__, var, prop))

    def check_parent_superset(cls, var):
        if not set(getattr(cls, var)).issuperset(set(getattr(cls.parent_class(), var))):
            raise TypeError("In '%s': the '%s' variable needs to be a superset of its class parent's one" % (cls.__name__, var))


    # validate that the schema is correctly defined
    check_present(cls, 'schema', dict, default_value =  False)

    # inherit schema from parent class
    schema = Schema(parent.schema) # make a copy of parent's schema
    schema.update(cls.schema) # make sure we don't overwrite (or do we want to allow overloading of variables?)
    schema._implicit = set(parent.schema._implicit) # no need to get cls.schema._implicit as well, it's empty
    cls.schema = schema

    # validate attribute types as defined in schema
    for name, ctype in cls.schema.items():
        if not isinstance(name, str) or not any(issubclass(ctype, dtype) for dtype in valid_types):
            raise TypeError("In '%s': the schema should be a dict of 'str' to either one of those accepted types (or a subclass of them): %s'" % (cls.__name__, ', '.join("'%s'" % c.__name__ for c in valid_types)))


    # all the properties defined as subclasses of BaseObject need to have an
    # associated reverseLookup entry
    check_present(cls, 'reverse_lookup', dict)
    orig_reverse_lookup = cls.reverse_lookup if 'reverse_lookup' in attrs else {}

    # inherit reverseLookup from parent
    rlookup = dict(parent.reverse_lookup)
    rlookup.update(cls.reverse_lookup)
    cls.reverse_lookup = rlookup

    # check that we have reverseLookup names for all needed properties
    object_props = [ name for name, ctype in cls.schema.items() if issubclass(ctype, BaseObject) and name not in cls.schema._implicit ]
    reverse_lookup = [ prop for prop in cls.reverse_lookup.keys() if prop not in cls.schema._implicit ]

    diff = set(reverse_lookup).symmetric_difference(set(object_props))
    if diff:
        raise TypeError("In '%s': you should define exactly one reverseLookup name for each property in your schema that is a subclass of BaseObject, different ones: %s" % (cls.__name__, ', '.join("'%s'" % c for c in diff)))

    # directly update the schema for other classes where needed
    # TODO: make sure we don't overwrite anything (should have been done in the validate_class_definition, right?)
    for prop, rprop in orig_reverse_lookup.items():
        for c in subclasses(cls.schema[prop]):
            c.schema._implicit.add(rprop)
            c.schema[rprop] = cls
            c.reverse_lookup[rprop] = prop


    # check that the other variables are correctly defined
    check_schema_subset(cls, 'valid', default_value = False)
    check_parent_superset(cls, 'valid')

    check_schema_subset(cls, 'unique')
    check_parent_superset(cls, 'unique')

    check_schema_subset(cls, 'order')
    # TODO: validate converters


def print_class(cls):
    print '*'*100
    print 'class: %s' % cls.__name__
    print 'parent: %s' % cls.parent_class().__name__
    print 'schema', cls.schema
    print 'implicit', cls.schema._implicit
    print 'rlookup', cls.reverse_lookup
    print '*'*100


def display_ontology():
    import tempfile
    import subprocess

    _, filename = tempfile.mkstemp(suffix = '.png')

    dg = []
    dg += [ 'digraph G {' ]

    #for _id, n in nodes.items():
    for cname, cls in _classes.items():
        label = '<FONT COLOR="#884444">%s</FONT><BR/>' % cname
        attrs = []
        for name, type in cls.schema.items():
            if name in cls.schema._implicit:
                continue
            attrs += [ '%s: %s' % (name, type.__name__) ]
        for name in cls.schema._implicit:
            attrs += [ '<FONT COLOR="#666666">%s: %s</FONT>' % (name, cls.schema[name].__name__) ]

        label += '<BR/>'.join(attrs)
        dg += [ 'node_%s [shape=polygon,sides=4,label=<%s>];' % (cname, label) ]

        dg += [ 'node_%s -> node_%s;' % (cname, cls.parent_class().__name__) ]


    dg += [ '}' ]

    subprocess.Popen([ 'dot', '-Tpng', '-o', filename ], stdin = subprocess.PIPE).communicate('\n'.join(dg))

    subprocess.Popen([ 'gwenview', filename ], stdout = subprocess.PIPE, stderr = subprocess.PIPE).communicate()

def register(cls, attrs):
    # when registering BaseObject, skip the tests
    if cls.__name__ == 'BaseObject':
        _classes['BaseObject'] = cls
        # save this as a default ontology, it might be useful
        save_current_ontology('origin')
        return

    if cls.__name__ in _classes:
        if cls == _classes[cls.__name__]:
            log.info('Class %s already registered' % cls.__name__)
            return

        log.warning('Found previous definition of class %s. Ignoring new definition...' % cls.__name__)
        return

    validate_class_definition(cls, attrs)

    _classes[cls.__name__] = cls

    revalidate_graphs()

    #displayOntology()


def register_graph(graph):
    _graphs[id(graph)] = graph

def get_class(class_name):
    """Returns the ObjectNode class object given its name."""
    try:
        return _classes[class_name]
    except:
        raise ValueError, 'Class "%s" has not been registered with the OntologyManager' % class_name

def class_names():
    return _classes.keys()

def import_class(cls):
    sys._getframe(1).f_globals[cls] = get_class(cls)

def import_classes(classes):
    """Import the given classes in the caller's global variables namespace."""
    for cls in classes:
        sys._getframe(1).f_globals[cls] = get_class(cls)

def import_all_classes():
    import_classes(class_names())

def import_ontology(name):
    reload_saved_ontology(name)
    import_all_classes()



from setuptools import setup, find_packages
import sys, os

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.txt')).read()
NEWS = open(os.path.join(here, 'NEWS.txt')).read()


version = '0.1a1'

install_requires = [
    # List your project dependencies here.
    # For more details, see:
    # http://packages.python.org/distribute/setuptools.html#declaring-dependencies
]


setup(name = 'pygoo',
      version = version,
      description = "PyGoo is an Object-Graph mapper, similar to SQLAlchemy but for graph DBs",
      long_description = README + '\n\n' + NEWS,
      classifiers = [], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords = 'pygoo object graph mapper',
      author = 'Nicolas Wack',
      author_email = 'wackou@gmail.com',
      url = 'http://gitorious.org/pygoo',
      license = 'GPLv3',
      packages = find_packages(exclude = [ 'ez_setup', 'examples', 'tests' ]),
      include_package_data = True,
      zip_safe = False,
      install_requires = install_requires,
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyGoo - An Object-Graph mapper
# Copyright (c) 2010 Nicolas Wack <wackou@gmail.com>
#
# PyGoo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyGoo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from pygoo.abstractnode import AbstractNode
from pygoo.abstractdirectedgraph import Equal
from pygoo.baseobject import BaseObject
from pygoo.utils import tolist, toresult, is_of, multi_is_instance, is_literal
from pygoo import ontology
import types
import weakref
import logging

log = logging.getLogger('pygoo.ObjectNode')



class ObjectNode(AbstractNode):
    """An ObjectNode is a nice and useful mix between an OOP object and a node in a graph.

    An ObjectNode behaves in the following way:
     - it can have any number of named properties, of any type (literal type or another ObjectNode)
         If the property is a literal type, it is stored inside the node
         If the property is another node(s), it means there are directed edges of the same name from this node to the other one(s)
     - it implements dotted attribute access.
     - it keeps a list of valid classes for this node. If a node has a certain class, we can then create a valid instance of
       that class (subclass of BaseObject) with the data from this node. The list of valid classes is actualized each time you
       set a property.

    ObjectNodes should implement different types of equalities:
      - identity: both refs point to the same node in the ObjectGraph
      - all their properties are equal (same type and values)
      - DEPRECATED(*) all their standard properties are equal
      - DEPRECATED(*) only their primary properties are equal

    (*) this should now be done in BaseObject instances instead of directly on the ObjectNode.

    ---------------------------------------------------------------------------------------------------------

    To be precise, ObjectNodes use a type system based on relaxed type classes
    (http://en.wikipedia.org/wiki/Type_classes)
    where there is a standard object hierarchy, but an ObjectNode can be of various distinct
    classes at the same time.

    As this doesn't fit exactly with python's way of doing things, class value should be tested
    using the ObjectNode.isinstance(class) method instead of the usual isinstance(obj, class) function.

    ---------------------------------------------------------------------------------------------------------

    Not yet implemented / TODO:

    Accessing properties should return a "smart" iterator when accessing properties which are instances of
    AbstractNode, which also allows to call dotted attribute access on it, so that this becomes possible:

    for f in Series.episodes.file.filename:
        do_sth()

    where Series.episodes returns multiple results, but Episode.file might also return multiple results.
    File.filename returns a literal property, which means that we can now convert our iterator over BaseObject
    into a list (or single element) of literal
    """

    def __init__(self, graph, props = []):
        super(ObjectNode, self).__init__(graph, props)
        log.debug('ObjectNode.__init__: props = %s' % str(props))

        self.graph = weakref.ref(graph)

        for prop, value, reverse_name in props:
            self.set(prop, value, reverse_name, validate = False)

        self.update_valid_classes()


    def is_valid_instance(self, cls):
        """Returns whether this node can be considered a valid instance of a class given its current properties.

        This method doesn't use the cached value, but does the actual checking of whether there is a match."""
        return self.has_valid_properties(cls, cls.valid)

    def has_valid_properties(self, cls, props):
        for prop in props:
            value = self.get(prop)

            if isinstance(value, types.GeneratorType):
                value = list(value)
                # TODO: we might need value.isValidInstance in some cases
                if value != [] and not value[0].isinstance(cls.schema[prop]):
                    return False
            else:
                # TODO: here we might want to check if value is None and allow it or not
                if type(value) != cls.schema[prop]:
                    return False

        return True


    def invalid_properties(self, cls):
        invalid = []
        for prop in cls.valid:
            if prop not in self:
                invalid.append("property '%s' is missing" % prop)
                continue

            # FIXME: link type checking doesn't work
            if isinstance(getattr(self, prop), types.GeneratorType):
                continue

            if not multi_is_instance(getattr(self, prop), cls.schema[prop]):
                invalid.append("property '%s' is of type '%s', but should be of type '%s'" %
                               (prop, type(getattr(self, prop)).__name__, cls.schema[prop].__name__))

        return '\n'.join(invalid)


    def update_valid_classes(self):
        """Revalidate all the classes for this node."""
        if self.graph()._dynamic:
            self.clear_classes()
            for cls in ontology._classes.values():
                if self.is_valid_instance(cls):
                    self.add_class(cls)
        else:
            # if we have static inheritance, we don't want to do anything here
            pass

        log.debug('valid classes for %s:\n  %s' % (self.to_string(), [ cls.__name__ for cls in self._classes ]))

    def virtual_class(self):
        """Return the most specialized class that this node is an instance of."""
        cls = BaseObject
        for c in self.classes():
            if issubclass(c, cls):
                cls = c
        return cls

    def virtual(self):
        """Return an instance of the most specialized class that this node is an instance of."""
        return self.virtual_class()(self)

    ### Container methods

    def keys(self):
        for k in self.literal_keys():
            yield k
        for k in self.edge_keys():
            yield k

    def values(self):
        for v in self.literal_values():
            yield v
        for v in self.edge_values():
            yield v

    def items(self):
        for i in self.literal_items():
            yield i
        for i in self.edge_items():
            yield i

    def __contains__(self, name):
        return name in self.keys()

    def __iter__(self):
        for prop in self.keys():
            yield prop

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        # TODO: why do we need this again?
        raise NotImplementedError


    ### Acessing properties methods

    def __getattr__(self, name):
        try:
            return self.get_literal(name)
        except:
            try:
                return self.outgoing_edge_endpoints(name) # this should be an iterable over the pointed nodes
            except:
                raise AttributeError(name)

    def get(self, name):
        """Returns the given property or None if not found.
        This can return either a literal value, or an iterator through other nodes if
        the given property actually was a link relation."""
        try:
            return self.__getattr__(name)
        except AttributeError:
            return None

    def get_chained_properties(self, prop_list):
        """Given a list of successive chained properties, returns the final value.
        e.g.: Movie('2001').getChainedProperties([ 'director', 'firstName' ]) == 'Stanley'

        In case some property does not exist, it will raise an AttributeError."""
        result = self
        for prop in prop_list:
            result = result.get(prop)
            if isinstance(result, types.GeneratorType):
                # FIXME: this will fail if it branches before the last property
                result = toresult(list(result))

        return result

    ### properties manipulation methods

    def __setattr__(self, name, value):
        if name in [ 'graph' ]:
            object.__setattr__(self, name, value)
        else:
            self.set(name, value)

    def set(self, name, value, reverse_name = None, validate = True):
        """Sets the property name to the given value.

        If value is an ObjectNode, we're actually setting a link between them two, so we use reverseName as the
        name of the link when followed in the other direction.
        If reverseName is not given, a default of 'isNameOf' (using the given name) will be used."""

        if multi_is_instance(value, AbstractNode):
            if reverse_name is None:
                reverse_name = is_of(name)

            self.set_link(name, value, reverse_name)

        elif is_literal(value):
            self.set_literal(name, value)

        else:
            raise TypeError("Trying to set property '%s' of %s to '%s', but it is not of a supported type (literal or object node): %s" % (name, self, value, type(value).__name__))

        # update the cache of valid classes
        if validate:
            self.update_valid_classes()



    def append(self, name, value, reverse_name = None, validate = True):
        if multi_is_instance(value, AbstractNode):
            if reverse_name is None:
                reverse_name = is_of(name)

            self.add_link(name, value, reverse_name)

        else:
            raise TypeError("Trying to append to property '%s' of %s: '%s', but it is not of a supported type (literal or object node): %s" % (name, self, value, type(value).__name__))

        # update the cache of valid classes
        if validate:
            self.update_valid_classes()


    def add_link(self, name, other_node, reverse_name):
        g = self.graph()

        if isinstance(other_node, list) or isinstance(other_node, types.GeneratorType):
            for n in other_node:
                g.add_link(self, name, n, reverse_name)
        else:
            g.add_link(self, name, other_node, reverse_name)


    def set_link(self, name, other_node, reverse_name):
        """Can assume that otherNode is always an object node or an iterable over them."""
        # need to check for whether otherNode is an iterable
        #if self._graph != otherNode._graph:
        #    raise ValueError('Both nodes do not live in the same graph, cannot link them together')

        g = self.graph()

        # first remove the old link(s)
        # Note: we need to wrap the generator into a list here because it looks like otherwise
        # the removeLink() call messes up with it
        for n in list(self.get(name) or []): # NB: 'or []' because if the property doesn't exist yet, self.get() returns None
            g.remove_link(self, name, n, reverse_name)

        # then add the new link(s)
        self.add_link(name, other_node, reverse_name)



    def update(self, props):
        """Update this ObjectNode properties with the ones contained in the given dict.
        Should also allow instances of other ObjectNodes."""
        for name, value in props.items():
            self.set(name, value, validate = False)

        self.update_valid_classes()

    def update_new(self, other):
        """Update this ObjectNode properties with the only other ones it doesn't have yet."""
        raise NotImplementedError

    def same_properties(self, other, props = None, exclude = [], cmp = None):
        # NB: sameValidProperties and sameUniqueProperties should be defined in BaseObject
        # TODO: this can surely be optimized
        if props is None:
            props = other.items()
        else:
            props = [ (p, other.get(p)) for p in props ]

        for name, value in props:
            if name in exclude:
                continue
            #print '     prop:', name,
            if isinstance(value, types.GeneratorType):
                svalue = list(self.get(name))
                value = list(value)
                #print 'gen; value=', svalue, value

                if cmp == Equal.OnUnique:
                    # FIXME: this is an ugly workaround, but I need to get pygoo back on track
                    #        and this has to work *right now*
                    def same_props_obj(obj1, obj2):
                        props = obj1.unique_properties()
                        return obj1.node.same_properties(obj2.node, props, cmp = Equal.OnUnique)

                    result = (len(svalue) == len(value) and
                              all(same_props_obj(v1.virtual(), v2.virtual()) for v1, v2 in zip(svalue, value)))
                else:
                    # FIXME: v1.virtual() should not be used here...
                    result = (len(svalue) == len(value) and
                              all(v1.virtual() == v2.virtual() for v1, v2 in zip(svalue, value)))

                if result is False:
                    return False
            else:
                #print 'normal; value=', self.get(name), value
                if self.get(name) != value:
                    return False

        return True


    ### String methods

    def __str__(self):
        return self.to_string().encode('utf-8')

    def __unicode__(self):
        return self.to_string()

    def __repr__(self):
        return str(self)


    def to_string(self, cls = None, default = None):
        # TODO: smarter stringize that guesses the class, should it always be there?
        cls = self.virtual_class()

        if cls is None:
            # most likely called from a node, but anyway we can't infer anything on the links so just display
            # them as anonymous ObjectNodes
            cls = self.__class__
            props = [ (prop, [ cls.__name__ ] * len(tolist(value))) if multi_is_instance(value, ObjectNode) else (prop, unicode(value)) for prop, value in self.items() ]

        else:
            props = []
            for prop, value in self.items():
                if prop in cls.schema._implicit:
                    continue
                elif isinstance(value, types.GeneratorType):
                    props.append((prop, unicode(toresult([ v.to_string(cls = cls.schema.get(prop) or default) for v in value ]))))
                else:
                    props.append((prop, unicode(value)))

        return u'%s(%s)' % (cls.__name__, ', '.join([ u'%s=%s' % (k, v) for k, v in props ]))
